#include "MusicManager.h"
#include "SDL.h"

string MusicManager::getTrackName() {
	if(curTrack < 0 || curTrack > (int)tracks.size()) {
		return string("Not playing music");
	}
	if(tracks[curTrack]->tags != NULL) {
		return string(tracks[curTrack]->tags->artist + " - " + tracks[curTrack]->tags->title);
	}
	return string("File not tagged");
}

void MusicManager::setVolume(int vol) {
	Mix_VolumeMusic(vol);
}

void MusicManager::pause() {
	Mix_PauseMusic();
}

void MusicManager::restart() {
	Mix_RewindMusic();
}

void MusicManager::resume() {
	Mix_ResumeMusic();
}

void MusicManager::changeTrack(int dt) {
	curTrack += dt;
	if(curTrack < 0) {
		curTrack = (int)tracks.size()-1;
	}
	if(curTrack >= (int)tracks.size()) {
		curTrack = 0;
	}
	Mix_HaltMusic();
	playTrack(curTrack, true);
}

void MusicManager::playTrack(int track, bool bLooping) {
	if(track < 0 || track >= (int)tracks.size()) {
		fprintf(stderr, "Err.) Music track %i is out of range [0,%i]\n", track, (int)tracks.size()-1);
		return;
	}
	
	int reps = 0;
	if(bLooping)
		reps = -1;
	
	if(Mix_PlayMusic(tracks[track]->data, reps) == -1) {
		fprintf(stderr, "Err.) SDL_mixer error %s\n", Mix_GetError());
		return;
	}
	printf("Pass) Playing track %i", track);
	if(tracks[track]->tags != NULL) {
		printf(" (%s - %s)\n", tracks[track]->tags->artist.c_str(), tracks[track]->tags->title.c_str());
	}
	else {
		printf("\n");
	}
	
	Mix_Volume(0, 100);
	
	curTrack = track;
}

MusicManager::~MusicManager() {
	Mix_HaltMusic();
	for(unsigned int i = 0; i < tracks.size(); i++) {
		Mix_FreeMusic(tracks[i]->data);
		delete tracks[i]->tags;
		delete tracks[i];
	}
	tracks.clear();
	Mix_CloseAudio();
}

MusicManager::MusicManager(char *dir) {
	curTrack = -1;
	char *filename = new char[400];
	int start = SDL_GetTicks();
	for(int i = 0; ; i++) {
		sprintf(filename, "%s/%i.ogg", dir, i);
		if(loadTrack(filename) < 0) {
			fprintf(stderr, "Err.) Music file %s is missing\n", filename);
			if(i == 0) {
				fprintf(stderr, "Err.) No music found, disabling\n");
				return;
			}
			return;
		}
	}
	int timer = SDL_GetTicks() - start;
	printf("Pass) Music loaded from %s in %ims.\n", dir, timer);
	delete[] filename;
}

int MusicManager::loadTrack(char *fname) {
	Mix_Music* thisTrack = Mix_LoadMUS(fname);
	if(!thisTrack) {
		fprintf(stderr, "Err.) Can't load %s... %s\n", fname, Mix_GetError());
		return -1;
	}
	// Now load the ID3 tags and shove them in
	MusicTrack* thisFinalTrack = new MusicTrack;
	thisFinalTrack->tags = getV1Tags(fname);
	thisFinalTrack->data = thisTrack;
	tracks.push_back(thisFinalTrack);
	return 0;
}
