#ifndef _PROPANEACTOR_H_
#define _PROPANEACTOR_H_

#include "SpriteManager.h"

/**
 * A 2D sprite based actor with simple animation and collision support.
 * Can be overridden for NetworkActor, Propane3DActor, etc.
 */

class PropaneActor {
public:
	PropaneActor(SpriteManager* sprites);					/// create with specified sprite manager
	virtual ~PropaneActor();
	void setAnimation(int startFrame, int numFrames, float frameDelay);	/// override default animation values permanently
	virtual void draw(int px, int py);					/// draw the sprite to screen using the sprite manager

	virtual void move(float dx, float dy);					/// move the sprite a certain amount of world units
	virtual void temporaryFrame(int frame);					/// override animation temporarily, for a single frame
	virtual void resetAnimation();						/// return animation to default values
	virtual void rotateTowards(double maxRot, double goal);			/// rotate to face a certain angle, useful for turning
	virtual void rotateFacing(float x, float y);				/// rotate to face a given (x,y) in world units

	float getX();								/// get the current x-coordinate in world units
	float getY();								/// get the current y-coordinate in world units
	float getWidth();							/// get the current width in world units
	float getHeight();							/// get the current height in world units
	float getForwardX(float delta);
	float getForwardY(float delta);
	virtual double getRotation();						/// get current rotation in degrees
	virtual double angleTowards(float wx, float wy);
	virtual double angleDistance(double angle1, double angle2);
	virtual double angleOutOf360(double angle);
	
	void setX(float newX);							/// set x-coordinate in WUs
	void setY(float newY);							/// set y-coordinate in WUs
	void setWidth(float width);						/// set render and collision width in WUs
	void setHeight(float height);						/// set render and collision height in WUs
	void setXYWH(float x, float y, float width, float height);
	virtual void setRotation(double degrees);				/// set current rotation in degrees
	virtual void rotate(double dRot);					/// change the current rotation by degrees

	bool collision(PropaneActor* actor2);					/// detect collision with another PropaneActor

	int getFrame();								/// return the sprite that should be used for rendering
	void updateAnimation(float delta);					/// update the animation, increment frames as necessary
protected:
	int startFrame;								/// the first sprite in the animation series
	float frameTimer;							/// the amount of time the current frame has spent
	float frameDelay;							/// the amount of time a frame can take before it is switched
	int curFrame;								/// the current offset from the first sprite in the series
	int numFrames;								/// how high the offset can get before it loops back to the first frame

	float x;								/// x coordinate, in world units (WU)
	float y;								/// y coordinate, WU
	float w;								/// render and collision width, WU
	float h;								/// render and collision height, WU
	double rotation;							/// the current rotation in degrees
	
	float DegToRad(double degree);						/// converts degrees to radians (slow)
	double rotationOutOf360();						/// converts the rotation to be limited to (0,360)

	bool inTemp;								/// are we in a temporary frame?
	int tempFrame;								/// the id of the temporary sprite

	SpriteManager* myMgr;							/// a pointer to the sprite manager to use to render this
};

#endif
