#include "RavBitmapFont.h"
#include "SDL_image.h"
#include "glheaders.h"
#include <fstream>
using namespace std;

void RavBitmapFont::print(string text, int x, int y) {
	if(!bIsReady) {
		fprintf(stderr, "FAIL) This font isn't ready to print to. Didn't complete construction\n");
		return;
	}
	int cursorX = x;
	int cursorY = y;
	for(unsigned int i = 0; i < text.size(); i++) {
		// Get the ID from the text.
		int thisOne = (char)text[i];
		if(thisOne == '\\') {
			cursorX = x;
			cursorY += getLineHeight();
		}
		
		// Make up the source rectangle to pull it from the image
		int srcLeft = characters[thisOne].x;
		int srcTop = characters[thisOne].y;
		int srcRight = characters[thisOne].width + srcLeft;
		int srcBottom = characters[thisOne].height + srcTop;
		
		// Make up the destination rectangle.
		int destLeft = cursorX + characters[thisOne].xoffset;
		int destTop = cursorY + characters[thisOne].yoffset;
		
		// Draw it somehow.
		drawChar(srcLeft, srcTop, srcRight, srcBottom, destLeft, destTop, characters[thisOne].width, characters[thisOne].height, r, g, b, 1.0f);
		
		// Update the cursor.
		cursorX += characters[thisOne].xadvance;
	}
}

void RavBitmapFont::sprint(int x, int y, string text, ...) {
	/*
		Courtesy of Laz (http://www.lameware.com)
		Nov. 12 2005
	*/
	if(!bIsReady) {
		fprintf(stderr, "FAIL) This font isn't ready to print to. Didn't complete construction\n");
		return;
	}
	int cursorX = x;
	int cursorY = y;

	// Set up the new string
	va_list ap;
	char tmpText[256];
	va_start(ap, text);
		vsprintf(tmpText, text.c_str(), ap);
	va_end(ap);
	print(tmpText, x, y);
}

void RavBitmapFont::print(string text, int x, int y, float alpha) {
	if(!bIsReady) {
		fprintf(stderr, "FAIL) This font isn't ready to print to. Didn't complete construction\n");
		return;
	}
	int cursorX = x;
	int cursorY = y;
	for(unsigned int i = 0; i < text.size(); i++) {
		// Get the ID from the text.
		int thisOne = (char)text[i];
		
		if(thisOne == '\\') {
			cursorX = x;
			cursorY += getLineHeight();
			continue;
		}
		// Make up the source rectangle to pull it from the image
		int srcLeft = characters[thisOne].x;
		int srcTop = characters[thisOne].y;
		int srcRight = characters[thisOne].width + srcLeft;
		int srcBottom = characters[thisOne].height + srcTop;
		
		// Make up the destination rectangle.
		int destLeft = cursorX + characters[thisOne].xoffset;
		int destTop = cursorY + characters[thisOne].yoffset;
		
		// Draw it somehow.
		drawChar(srcLeft, srcTop, srcRight, srcBottom, destLeft, destTop, characters[thisOne].width, characters[thisOne].height, r, g, b, alpha);
		
		// Update the cursor.
		cursorX += characters[thisOne].xadvance;
	}
	//glColor3f(1.0f, 1.0f, 1.0f);
}

void RavBitmapFont::rightAlignPrint(string text, int x, int y, float alpha) {
	if(!bIsReady) {
		fprintf(stderr, "FAIL) This font isn't ready to print to. Didn't complete construction\n");
		return;
	}
	int cursorX = x;
	int cursorY = y;
	for(int i = text.size() - 1; i >= 0; i--) {
		// Get the ID from the text.
		int thisOne = (char)text[i];
		
		if(thisOne == '\\') {
			cursorX = x;
			cursorY += getLineHeight();
			continue;
		}
		// Make up the source rectangle to pull it from the image
		int srcLeft = characters[thisOne].x;
		int srcTop = characters[thisOne].y;
		int srcRight = characters[thisOne].width + srcLeft;
		int srcBottom = characters[thisOne].height + srcTop;
		
		// Make up the destination rectangle.
		int destLeft = (cursorX - characters[thisOne].width) + characters[thisOne].xoffset;
		int destTop = cursorY + characters[thisOne].yoffset;
		
		// Draw it somehow.
		drawChar(srcLeft, srcTop, srcRight, srcBottom, destLeft, destTop, characters[thisOne].width, characters[thisOne].height, r, g, b, alpha);
		
		// Update the cursor.
		cursorX -= characters[thisOne].xadvance;
	}
	//glColor3f(1.0f, 1.0f, 1.0f);
}

int getInt(string tok) {
	string usetok = "";
	for(unsigned int i = 0; i < tok.size(); i++) {
		if(tok[i] != '\n' && tok[i] != '\r' && tok[i] != ' ') {
			usetok += tok[i];
		}
	}
#if defined(RAV_DEBUG)
	printf("INFO) integering token '%s'\n", usetok.c_str());
#endif
	int siz = usetok.size();
	if(siz < 1) {
		return -1;
	}
	char *spr = new char[siz+1];
	strcpy(spr, usetok.c_str());
	//char *name = strsep(&spr, "=");
	char *name = strtok(spr, "=");
	char *result = strtok(NULL, "\0");
	if(strlen(name) < 1) {
		fprintf(stderr, "WARN) Name too short: '%s' (must be at least 1 character)\n", name);
	}
	int num = atoi(result);
#if defined(RAV_DEBUG)
	printf("INFO) Name: %s Value: %i\n", name, num);
#endif
	//delete[] name;
	//delete[] spr;
	return num;
}

int RavBitmapFont::getLineHeight() { 
	return lineHeight;
}

RavBitChar RavBitmapFont::generateCharacterFromLine(char *line) {
	if(strlen(line) < 1) {
		fprintf(stderr, "WARN) Detected blank line.\n");
		RavBitChar dummy;
		dummy.char_id = -66;
		return dummy;
	}
	RavBitChar newChar;
	
	bool bDone = false;
	vector <string> tokens;
#if defined(RAV_DEBUG)
	printf("INFO) Collecting tokens from \"%s\"\n", line);
#endif
	bool bFirstTime = true;
	while(!bDone) {
		char *foo = NULL;
		if(bFirstTime) {
			foo = strtok(line, " ");
			bFirstTime = false;
		}
		else {
			foo = strtok(NULL, " ");
		}
		if(foo != NULL) {
			tokens.push_back(foo);
		}
		else {
			bDone = true;
		}
	}
#if defined(RAV_DEBUG)
	printf("INFO) Interpreting tokens (%i tokens found)\n", (int)tokens.size());
#endif
	if(tokens.size() < 9) {
		fprintf(stderr, "FAIL) Line \"%s\" does not contain enough tokens to be a character ID\n", line);
		if(tokens[0].compare("kerning") == 0) {
			newChar.char_id = -2;
#if defined(RAV_DEBUG)
			printf("INFO) This is a kerning line, just breaking. We're done\n");
#endif
			return newChar;
		}
		exit(1);
	}
	if(tokens[0].compare("char") != 0) {
		fprintf(stderr, "FAIL %s:%i) Line \"%s\" does not contain magic word 'char' at start. Not character id.\n", __FILE__, __LINE__, line);
		exit(1);
	}
	// Remove all blank tokens
	vector <string> tempVec;
	for(unsigned int i = 0; i < tokens.size(); i++) {
		if(tokens[i].length() > 1) {
			tempVec.push_back(tokens[i]);
		}
	}
	tokens.clear();
	for(unsigned int i = 0; i < tempVec.size(); i++) {
		tokens.push_back(tempVec[i]);
	}
	tempVec.clear();
	for(unsigned int i = 0; i < tokens.size(); i++) {
#if defined(RAV_DEBUG)
		printf("INFO) Processing token '%s'\n", tokens[i].c_str());
#endif
		// Sometimes there are zero-size tokens! WTF!
		char *thisTok = (char*)tokens[i].c_str();
		
		// What should I do here? I'm using a struct, not a map
		switch(i) {
			case 0:
				// 'char'
				break;
			case 1:
				// id
				newChar.char_id = getInt(thisTok);
				break;
			case 2:
				// x
				newChar.x = getInt(thisTok);
				break;
			case 3:
				// y
				newChar.y = getInt(thisTok);
				break;
			case 4:
				// width
				newChar.width = getInt(thisTok);
				break;
			case 5:
				// height
				newChar.height = getInt(thisTok);
				break;
			case 6:
				// xoff
				newChar.xoffset = getInt(thisTok);
				break;
			case 7:
				// yoff
				newChar.yoffset = getInt(thisTok);
				break;
			case 8:
				// xadvance
				newChar.xadvance = getInt(thisTok);
				break;
			case 9:
				// page.
				newChar.page = getInt(thisTok);
				break;
			default:
				fprintf(stderr, "FAIL) Too many tokens in this line.\n");
				break;
		}
	}
	// Do whatever cleanup you have to
	return newChar;
}

void RavBitmapFont::drawChar(int srcX, int srcY, int srcR, int srcB, int destX, int destY, int destW, int destH,
							 Uint8 r, Uint8 g, Uint8 b, float af) {
	// Bind texture
	// Colours
	float rf = (float)(r/255.0f);
	float gf = (float)(g/255.0f);
	float bf = (float)(b/255.0f);
	// Texture coordinates -- these divides are probably fucking killing us
	float sx = (float)(srcX / 256.0f);
	float sy = (float)(srcY / 256.0f);
	float sr = (float)(srcR / 256.0f);
	float sb = (float)(srcB / 256.0f);
	
	// Draw the quad.
	glPushMatrix();
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	//glRotatef(180, 0, 0, 1);
	glBindTexture(GL_TEXTURE_2D, *texturePages[0]);
	if(texturePages[0] == NULL) {
		fprintf(stderr, "Page #0 is gone!\n");
		exit(1);
	}
	
	glBegin(GL_QUADS);
	// Set the bitmap colour
	glColor4f(rf, gf, bf, af);
	// Top left
	glTexCoord2f(sx, sy);
	//glTexCoord2f(0.0f, 0.0f);
	glVertex2i(destX, destY);
	// Top right
	glTexCoord2f(sr, sy);
	//glTexCoord2f(1.0f, 0.0f);
	glVertex2i(destX + destW, destY);
	// Bottom right
	glTexCoord2f(sr, sb);
	//glTexCoord2f(1.0f, 1.0f);
	glVertex2i(destX + destW, destY + destH);
	// Bottom left
	glTexCoord2f(sx, sb);
	//glTexCoord2f(0.0f, 1.0f);
	glVertex2i(destX, destY + destH);
	glEnd();
	
	// Make sure we didn't break anything. Christ!
	glColor3f(1.0f, 1.0f, 1.0f);
	glPopMatrix();
}

void RavBitmapFont::setEnvironmentVars(char *line) {
	bool bDone = false;
	vector <string> tokens;
#if defined(RAV_DEBUG)
	printf("INFO) Collecting tokens from \"%s\"\n", line);
#endif
	bool bFirstTime = true;
	while(!bDone) {
		char *foo = NULL;
		if(bFirstTime) {
			foo = strtok(line, " ");
			bFirstTime = false;
		}
		else {
			foo = strtok(NULL, " ");
		}
		if(foo != NULL) {
			tokens.push_back(foo);
		}
		else {
			bDone = true;
		}
	}
#if defined(RAV_DEBUG)
	printf("INFO) Interpreting tokens (%i tokens found)\n", (int)tokens.size());
#endif
	if(tokens.size() < 6) {
		fprintf(stderr, "FAIL) Line \"%s\" does not contain enough tokens to be a commons definition\n", line);
		exit(1);
	}
	if(tokens[0].compare("common") != 0) {
		fprintf(stderr, "FAIL) Line \"%s\" does not contain magic word 'common' at start. Not commons definition.\n", line);
		exit(1);
	}
	for(unsigned int i = 0; i < tokens.size(); i++) {
#if defined(RAV_DEBUG)
		printf("INFO) Processing token '%s'\n", tokens[i].c_str());
#endif
		char *thisTok = (char*)tokens[i].c_str();
		
		// What should I do here? I'm using a struct, not a map
		switch(i) {
			case 0:
				// common
				break;
			case 1:
				// lineHeight
				printf("setting lineHeight\n");
				lineHeight = getInt(thisTok);
				break;
			case 2:
				// base
				printf("setting base\n");
				base = getInt(thisTok);
				break;
			case 3:
				// scaleW
				printf("setting scaleW\n");
				scaleW = getInt(thisTok);
				break;
			case 4:
				// scaleH
				printf("setting scaleH\n");
				scaleH = getInt(thisTok);
				break;
			case 5:
				// pages
				printf("ignoring pages for now\n");
				//pages = getInt(thisTok);
				break;
			default:
				fprintf(stderr, "FAIL) Too many tokens in this line.\n");
				break;
		}
	}
	// Do whatever cleanup you have to
}

int RavBitmapFont::loadDefinition(char *filename) {
	ifstream TheFile(filename);
	if(!TheFile) {
		fprintf(stderr, "[%s] wasn't found\n", filename);
		return -1;
	}
	
	char infoLine[202];
	TheFile.getline(infoLine, 200);
	char commonLine[202];
	TheFile.getline(commonLine, 200);
	setEnvironmentVars(commonLine);
	
	while(!TheFile.eof()) {
		// First we'll get the common info.
		char* line = new char[202];
		TheFile.getline(line, 200);
		//printf("Line is '%s'\n", line);
		RavBitChar newChar = generateCharacterFromLine(line);
		if(newChar.char_id > 0) {
			characters[newChar.char_id] = newChar;
		}
		delete[] line;
		if(newChar.char_id == -2) {
			return 0;
		}
	}
	return 0;
}

void RavBitmapFont::setFontColour(Uint8 mR, Uint8 mG, Uint8 mB) {
	r = mR;
	g = mG;
	b = mB;
}

int RavBitmapFont::loadBitmap(char *filename) {
	SDL_Surface *surf = IMG_Load(filename);
	if(!surf) {
		fprintf(stderr, "FAIL) Image file %s not found\n", filename);
		return -1;
	}
	// Cut out the black from it...
	SDL_SetColorKey(surf, SDL_SRCCOLORKEY, SDL_MapRGB(surf->format, 0, 0, 0));
	
	// Otherwise, convert it.
	//printf("INFO) Creating new GLuint\n");
	GLuint* m_content = new GLuint;
	
	//printf("INFO) glGenTextures\n");
	glGenTextures(1, m_content);
	
	//printf("INFO) Set Alpha\n");
	SDL_SetAlpha(surf, 0, 255);
	
	// thanks shilbert
	//printf("INFO) Endian trash\n");
	SDL_Surface* destImage = SDL_CreateRGBSurface(SDL_SWSURFACE,
												  surf->w,
												  surf->h,
												  32,
												  SDL_SwapLE32(0x000000FF),
												  SDL_SwapLE32(0x0000FF00),
												  SDL_SwapLE32(0x00FF0000),
												  SDL_SwapLE32(0xFF000000));
	
	//printf("INFO) Blitting\n");
	SDL_BlitSurface(surf, NULL, destImage, NULL);
	//printf("INFO) Binding\n");
	glBindTexture(GL_TEXTURE_2D, *m_content);
	//printf("INFO) Filter MIN\n");
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	//printf("INFO) Filter MAX\n");
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	//printf("INFO) Building Mip maps\n");
	SDL_LockSurface(destImage);
	gluBuild2DMipmaps(GL_TEXTURE_2D, 4, destImage->w, destImage->h, GL_RGBA, GL_UNSIGNED_BYTE, destImage->pixels);
	SDL_UnlockSurface(destImage);
	//printf("INFO) Freeing original surface\n");
	SDL_FreeSurface(surf);
	//printf("INFO) Freeing destination surface\n");
	SDL_FreeSurface(destImage);
	
	//printf("INFO) Storing converted surface.\n");
	texturePages.push_back(m_content);
	return 0;
}

RavBitmapFont::~RavBitmapFont() {
	printf("INFO) Emptying bitmap font\n");
	int start = SDL_GetTicks();
	for(unsigned int i = 0; i < texturePages.size(); i++) {
		delete texturePages[i];
	}
	texturePages.clear();
	int timer = SDL_GetTicks() - start;
	printf("PASS) Bitmap font emptied in %i ms\n", timer);
}

RavBitmapFont::RavBitmapFont(char *tgaFile, char *defFile) {
	r = 255;
	g = 255;
	b = 255;
	printf("INFO) Initializing bitmap font load (tga:%s fnt:%s)\n", tgaFile, defFile);
	int start = SDL_GetTicks();
	if(loadBitmap(tgaFile) < 0) {
		fprintf(stderr, "FAIL) Bitmap font bitmap file %s not found or unusable\n", tgaFile);
		return;
	}
	if(loadDefinition(defFile) < 0) {
		fprintf(stderr, "FAIL) Bitmap font definition file %s not found or unusable\n", defFile);
		return;
	}
	int timer = SDL_GetTicks() - start;
	printf("PASS) Startup sequence completed in %i ms using %s, %s\n", timer, tgaFile, defFile);
	bIsReady = true;
}

void RavBitmapFont::rightAlignPrint(string text, int x, int y) {
	rightAlignPrint(text, x, y, 1.0f);
}

RavBitmapFont::RavBitmapFont(char *commonName) {
	r = 255;
	g = 255;
	b = 255;
	printf("INFO) Initializing bitmap font load (family: %s)\n", commonName);
	int start = SDL_GetTicks();
	
	char *tgaFile = new char[100];
	sprintf(tgaFile, "%s.tga", commonName);
	if(loadBitmap(tgaFile) < 0) {
		fprintf(stderr, "FAIL) Bitmap font bitmap file %s not found or unusable\n", tgaFile);
		delete[] tgaFile;
		return;
	}
	delete[] tgaFile;
	
	char *defFile = new char[100];
	sprintf(defFile, "%s.fnt", commonName);
	
	if(loadDefinition(defFile) < 0) {
		fprintf(stderr, "FAIL) Bitmap font definition file %s not found or unusable\n", defFile);
		delete[] defFile;
		return;
	}
	delete[] defFile;
	
	int timer = SDL_GetTicks() - start;
	printf("PASS) Bitmap font load completed in %i ms\n", timer);
	bIsReady = true;	
}
