/*
	RavBitmapFont.h
 
	bitmap font support for openGL.
 
	works with angelcode bitmap fonts (www.angelcode.com/products/bmfont)
*/

#include "glheaders.h"
#include <string>
#include <vector>
using namespace std;

#ifndef _RAVBITMAPFONT_H_
#define _RAVBITMAPFONT_H_

#define NUM_CHARS 256

struct RavBitChar {
	int x;
	int y;
	int width;
	int height;
	int xadvance;
	int char_id;
	int page;
	int xoffset;
	int yoffset;
};

class RavBitmapFont {
public:
	RavBitmapFont(char *tgafilename, char *defFilename);
	RavBitmapFont(char *sharedfilename);
	
	virtual void print(string text, int x, int y);
	virtual void sprint(int x, int y, string text, ...);
	virtual void rightAlignPrint(string text, int x, int y, float alpha);
	virtual void rightAlignPrint(string text, int x, int y);
	virtual void print(string text, int x, int y, float alpha);
	void setFontColour(Uint8 r, Uint8 g, Uint8 b);
	virtual int getLineHeight();
	void debugPrint();
	
	virtual ~RavBitmapFont();
private:
	int loadBitmap(char *filename);
	int loadDefinition(char *filename);
	RavBitChar generateCharacterFromLine(char *line);
	void drawChar(int sx, int sy, int sr, int sb, int dx, int dy, int dw, int dh,
  Uint8 r, Uint8 g, Uint8 b, float alpha);
	void setEnvironmentVars(char *line);
	
	int lineHeight;
	int base;
	int scaleW;
	int scaleH;
	bool bIsReady;
	
	Uint8 r, g, b;
	
	vector <GLuint*> texturePages;
	RavBitChar characters[NUM_CHARS];
};

#endif
