#include "SoundEffectManager.h"
#include "SDL_mixer.h"

SoundEffectManager::SoundEffectManager(char *dir) {
	char *fname = new char[200];
	
	bool bDone = false;
	int i = 0;
	while(!bDone) {
		sprintf(fname, "%s/%i.wav", dir, i);
		if(addSample(fname) < 0) {
			fprintf(stderr, "%s is missing\n", fname);
			bDone = true;
		}
		i++;
	}
	
	delete[] fname;
}

void SoundEffectManager::setVolume(int vol) {
	Mix_Volume(-1, vol);
}

void SoundEffectManager::playSample(unsigned int samp) {
	// Compatibility purposes ONLY
	playSample(samp, false);
}

void SoundEffectManager::playSample(unsigned int samp, bool override) {
	if(samp < 0 || samp >= samples.size()) {
		printf("Sample %i is out of bounds\n", samp);
		return;
	}
	int chan = -1;
	if(override) {
		chan = 0;
	}
	int channel = Mix_PlayChannel(chan, samples[samp], 0);
	if(channel < 0) {
		fprintf(stderr, "Couldn't play sample %i: %s\n", samp, Mix_GetError());
	}
}

void SoundEffectManager::playSample(unsigned int samp, bool override, unsigned int overrideChan) {
	if(samp < 0 || samp >= samples.size()) {
		printf("Sample %i is out of bounds\n", samp);
		return;
	}
	int chan = -1;
	if(override) {
		chan = overrideChan;
	}
	int channel = Mix_PlayChannel(chan, samples[samp], 0);
	if(channel < 0) {
		fprintf(stderr, "Couldn't play sample %i: %s\n", samp, Mix_GetError());
	}
}


int SoundEffectManager::addSample(char *filename) {
	Mix_Chunk* chunk = Mix_LoadWAV(filename);
	if(!chunk) {
		return -1;
	}
	samples.push_back(chunk);
	return (int)(samples.size() - 1);
}
